from .make_dynamic_quantize_linear import (
    make_dynamic_quantize_linear_function_proto,
    make_simple_dynamic_quantize_linear_function_proto,
)
from .make_reshape_transpose import (
    make_matmul_reshape_transpose_function_proto,
    make_matmul_reshape_transpose_back_function_proto,
)
